/*
IAX balancing proxy
Authors: Ben Smithrust, Stanislaw Pitucha

Copyright (c) 2007 Gradwell dot com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "iax.h"
#include "clients.h"
#include "connections.h"
#include "global.h"
#include "list.h"
#include "servers.h"
#include "log.h"

static const char rcsid[] = "iaxproxy-version: " __FILE__ " " VCS_VERSION;

const char *close_types[] = {
	"Unexpected close by server",
	"Closed by server",
	"Unexpected close by client",
	"Closed by client",
	"Internal connection close"
};

// global connections list - used for cleanup
static list_ptrs global_connections_list = LIST_HEAD;

// list of acks that we will receive
static list_ptrs ack_wait_list = LIST_HEAD;
// list of pongs that we will receive and have to ack
//static list_ptrs pong_wait_list = LIST_HEAD;

// list of packets which may need retransmitting
static list_ptrs retransmissions_list = LIST_HEAD;

#define CALLNO_SIZE 32768
#define CALLNO_MIN 1
#define CALLNO_MAX (CALLNO_SIZE - 1)
#define CALLNO_TOTAL (CALLNO_MAX - CALLNO_MIN + 1)

// list of callno-s - as a list for O(1) geting first free one
static list_ptrs free_callno_list = LIST_HEAD;
// and as a table for O(1) lookup - it's just 32k*8 after all - anyone can handle that
static free_callno free_callnos[CALLNO_SIZE] = {{0,LIST_HEAD}};

static connection* new_connection(void) {
	connection *conn;

	conn = malloc(sizeof(connection));
	assert(conn!=NULL);

	conn->is_closing = 0;
	conn->closing = 0;
	conn->last_packet = 0;
	conn->setup_timestamp = 0;

	conn->from = NULL;
	conn->to = NULL;
	conn->user = NULL;

	conn->from_len = 0;
	conn->to_len = 0;
	conn->user_len = 0;
	
	list_init(&conn->retransmissions_list);

	return conn;
}

connection* connection_create_from_balancer( server *srv) {
	connection *conn;
	int32_t no;

	conn = new_connection();

	conn->srvno = 0;
	no = free_callno_get();
	if(no==-1) {
		LOG(ERROR, "failed to get a callno");
		return NULL;
	}

	conn->balancer_cltno = (uint16_t) no;
	conn->cltno = (uint16_t) no;
	LOG(DEBUG, "assigned own cltno - %i", conn->balancer_cltno);

	(void) time(&conn->created);
	conn->clt = NULL;
	conn->srv = srv;

	list_add_after(&conn->srv->opening_conns_head, &conn->opening_list);
	list_add_after(&global_connections_list, &conn->global_list);

	LOG(NOTICE, "new connection balancer->server at %s", srv->ip_string);

	return conn;
}

/*@null@*/ static connection* connection_create_balanced(client *clt, uint16_t cltno, uint16_t srvno) {
	connection *conn;
	int32_t no;

	conn = new_connection();

	conn->cltno = cltno;
	conn->srvno = srvno;
	no = client_get_callno(clt);
	if(no==-1) {
		LOG(ERROR, "failed to get a callno");
		return NULL;
	}

	conn->balancer_cltno = (uint16_t) no;
	LOG(DEBUG, "assigned balancer_cltno - %i", conn->balancer_cltno);

	(void) time(&conn->created);
	conn->clt = clt;
	conn->srv = server_get_best();
	if(conn->srv==NULL) {
		LOG(ERROR, "*** NO SERVERS TO HANDLE CONNECTION ***");
		client_return_callno(clt, no);
		return NULL;
	}
	conn->srv->load++;

	conn->clt->connections[conn->cltno] = conn;
	list_add_after(&conn->srv->opening_conns_head, &conn->opening_list);
	list_add_after(&global_connections_list, &conn->global_list);

	client_ref_up(clt); // count this as a new connection for client

	LOG_CONN(NOTICE, "new", conn);

	return conn;
}

/*@null@*/ connection* connection_get_by_clt(client *clt, uint16_t cltno, uint16_t srvno, char create) {
	connection *conn;

	conn = clt->connections[cltno];
	if(conn!=NULL) {
		LOG_CONN(VERBOSE_DEBUG, "found by clt", conn);
		return conn;
	}

	// no connection yet - create a new one if needed
	if(create==0)
		conn = NULL;
	else {
		conn = connection_create_balanced(clt, cltno, srvno);
	}
	return conn;
}

/*@null@*/ connection* connection_get_by_srv(server *srv, uint16_t cltno, uint16_t srvno) {
	list_ptrs *iter;
	connection *conn;

	conn = srv->connections[srvno];

	if(conn == NULL) {
		// second packet in connection probably
		int found=0;

		if(srvno==0) {
			LOG(ERROR, "INVAL answer to NEW, or something bizare - dropping");
			return NULL;
		}

		list_foreach(&srv->opening_conns_head, iter) {
			conn = list_entry(iter, connection, opening_list);

			if(conn->balancer_cltno == cltno) {
				conn->srvno = srvno;
				conn->srv->connections[srvno] = conn;

				LOG(NOTICE, "updated conn: server: %s clt: %i with srv: %i",
					conn->srv->ip_string, cltno, srvno);

				if(conn->clt!=NULL) { // isn't internal
					CLOG("Started call from %s/%i handled by %s/%i (balancer /%i), username `%.*s` from `%.*s` to `%.*s`.",
						conn->clt->ip_port, conn->cltno,
						conn->srv->ip_string, conn->srvno,
						conn->balancer_cltno,
						conn->user_len, conn->user,
						conn->from_len, conn->from,
						conn->to_len, conn->to);
				}

				list_remove(&conn->opening_list);
				found=1;
				break;
			}
		}

		// there was no half-open connection match
		// server doesn't start calls through proxy
		if(found==0) {
			// TODO: return INVAL packet
			LOG(ERROR, "first packet from server %s - this cannot happen", srv->ip_string);

			return NULL;
		}
	}

	// after previous block it's conn=connections[srv->cltno] is guaranteed
	// to exist with both srvno and cltno

	LOG_CONN(VERBOSE_DEBUG, "found by srv", conn);

	return conn;
}

void connection_cleanup_check(/*@only@*/ connection *conn, time_t now) {
	list_ptrs *iter, *next_iter;

	// check connections
	if ((conn->last_packet + CONN_TIMEOUT < now && conn->clt!=NULL) || // normal timeout
					(conn->last_packet + POKE_MAXWAIT < now && conn->clt==NULL) || // poke timeout
					(now < CLOSE_CAUSE_LAST) // forced end (hangup / inval)
					) {
		// if now==0 there was hangup, else it was timeout
		if(now>=5) {
			LOG_CONN(WARNING, "deleting timed out connection", conn);
		} else {
			LOG_CONN(NOTICE, "call finished - code: %s", conn, close_types[now]);
		}
		if(conn->clt!=NULL) {
			int length = conn->last_packet - conn->created;
			CLOG("Call from %s/%i handled by %s/%i (balancer /%i) finished after %02i:%02i (routed in %d.%03ds) - '%s'.",
				conn->clt->ip_port, conn->cltno,
				conn->srv->ip_string, conn->srvno,
				conn->balancer_cltno,
				length/60, length%60,
				conn->setup_timestamp/1000, conn->setup_timestamp%1000,
				close_types[(now<CLOSE_CAUSE_LAST)?now:CLOSED_INTERNAL_IGNORE]);
		}

		// remove all retransmissions
		list_foreach_safe(&conn->retransmissions_list, next_iter, iter)
			retransmission_free(list_entry(iter, retransmission, conn_list));

		// remove load only if it was clt->srv connection
		if(conn->clt)
			conn->srv->load--;

		if(conn->srv!=NULL && conn->srvno!=0) // if was answered
			conn->srv->connections[conn->srvno] = NULL;
		if(conn->clt!=NULL && conn->cltno!=0) // just in case...
			conn->clt->connections[conn->cltno] = NULL;

		if(list_is_linked(&conn->opening_list))
			list_remove(&conn->opening_list);
		list_remove(&conn->global_list);
		if(conn->balancer_cltno!=0) {
			if (conn->clt)
				client_return_callno(conn->clt, conn->balancer_cltno);
			else
				free_callno_return(conn->balancer_cltno);
		}

		if(conn->clt)
			client_ref_down_destroy(conn->clt);

		if(conn->from)
			free(conn->from);
		if(conn->to)
			free(conn->to);
		if(conn->user)
			free(conn->user);

		free(conn);
	}
}

void connection_cleanup_lists(struct timeval *now) {
	list_ptrs *iter, *next_iter;

	// check connections
	list_foreach_safe(&global_connections_list, next_iter, iter) {
		connection *conn;
		conn = list_entry(iter, connection, global_list);
		connection_cleanup_check(conn, now->tv_sec);
	}

	// check awaiting acks
	list_foreach_safe(&ack_wait_list, next_iter, iter) {
		ack_wait *ack;
		ack = list_entry(iter, ack_wait, list);
		if(ack->last_packet + CONN_TIMEOUT < now->tv_sec) {
			LOG(WARNING, "deleting timed out poke-ack wait from %s", ack->clt->ip_port);
			free_callno_return(ack->srvno);
			client_ref_down_destroy(ack->clt);

			list_remove(iter);
			free(ack);
		}
	}
}

void connection_destroy_all(void) {
	list_ptrs *iter;
	int destroyed = 0;
	while((iter = global_connections_list.next)!=NULL) {
		connection *conn;
		conn = list_entry(iter, connection, global_list);
		connection_cleanup_check(conn, 0);
		destroyed++;
	}
	LOG(NOTICE, "destroyed %i connections", destroyed);
}

void connection_print(connection *conn) {
	if(conn->clt) {
		fprintf(LOG_FILE, "connection server: %s:%i/%i client: %s/%i balancer: call %i\n",
			conn->srv->ip_string, ntohs(conn->srv->addr.sin_port), conn->srvno,
			conn->clt->ip_port, conn->cltno,
			conn->balancer_cltno);
	} else {
		fprintf(LOG_FILE, "connection server: %s:%i/%i balancer: call %i\n",
			conn->srv->ip_string, ntohs(conn->srv->addr.sin_port), conn->srvno,
			conn->balancer_cltno);
	}
}

void connection_list_print(void) {
	list_ptrs *srv_iter, *conn_iter;

	fprintf(LOG_FILE, "connections list:\n");
	list_foreach(&servers, srv_iter) {
		server *srv = list_entry(srv_iter, server, list);
		if(srv->enabled) {
			if(srv->not_responsive)
				fprintf(LOG_FILE, "- server ip=%s, load=%i (not responding)\n", srv->ip_string, srv->load);
			else
				fprintf(LOG_FILE, "- server ip=%s, load=%i\n", srv->ip_string, srv->load);
		} else
			fprintf(LOG_FILE, "- server ip=%s, load=%i (disabled)\n", srv->ip_string, srv->load);
		LOG_DO(DEBUG) {
			if(srv->opening_conns_head.next) {
				fprintf(LOG_FILE, "opening:\n");
				list_foreach(&srv->opening_conns_head, conn_iter) {
					connection *conn = list_entry(conn_iter, connection, opening_list);
					fprintf(LOG_FILE, "- ");
					connection_print(conn);
				}
				fprintf(LOG_FILE, "(end)\n\n");
			}
		}
	}
	LOG_DO(DEBUG) {
		list_foreach(&global_connections_list, conn_iter) {
			connection *conn = list_entry(conn_iter, connection, global_list);
			fprintf(LOG_FILE, "- ");
			connection_print(conn);
		}
	}
	fprintf(LOG_FILE, "(end)\n\n");
}

// this will return 1 if there's at least one active connection
// but result is not reliable as soon as function exits if there's >1 thread
int connection_check_if_exist(void) {
	list_ptrs *iter;

	list_foreach(&servers, iter) { // check each server for connections
		server *srv = list_entry(iter, server, list);
		if(srv->load>0) {
			return 1;
		}
	}
	return 0;
}

int connection_count(void) {
	int cnt = 0;
	list_ptrs *iter;
	
	list_foreach(&global_connections_list, iter)
		cnt++;

	return cnt;
}

int ack_wait_lookup(client *clt, uint16_t cltno, uint16_t srvno, uint32_t ts) {
	// return 1 if ack is for balancer + unlink, 0 otherwise
	list_ptrs *iter;
	ack_wait *ack;

	list_foreach(&ack_wait_list, iter) {
		ack = list_entry(iter, ack_wait, list);
		if(ack->clt==clt && ack->cltno==cltno && ack->srvno==srvno && ack->ts==ts) {
			list_remove(iter);
			free(ack);
			return 1;
		}
	}
	return 0;
}

void ack_wait_add(client *clt, uint16_t cltno, uint16_t srvno, uint32_t ts) {
	/*@only@*/ ack_wait *ack;

	ack = (ack_wait*) malloc(sizeof(ack_wait));
	assert(ack!=NULL);

	ack->clt = clt;
	ack->cltno = cltno;
	ack->srvno = srvno;
	ack->ts = ts;
	(void) time(&ack->last_packet);

	list_add_after(&ack_wait_list, &ack->list);
}

void free_callno_init(void) {
	int i;
	for (i = CALLNO_MIN; i <= CALLNO_MAX; i++) { // 0 is not used
		free_callnos[i].no=(uint16_t)i;
		list_add_after(&free_callno_list, &free_callnos[i].list);
	}
	free_callno_log();
}

static int free_callno_count(void) {
	int i = 0;
	list_ptrs *iter;
	list_foreach (&free_callno_list, iter)
		i++;
	if (i > CALLNO_TOTAL)
		CRASH("found %d free callno's - more than max total %d", i, CALLNO_TOTAL);
	return i;
}

/*
 * this code is quite slow but it is designed to run infrequently,
 * perhaps even not run at all in production code.
 */
void free_callno_log(void) {
	char used[CALLNO_SIZE], unused[CALLNO_SIZE];
	list_ptrs *iter, *conn_iter;
	int i, j, no;
	connection *conn;
	server *srv;

	i = free_callno_count();
	LOG(DEBUG, "Currently have %d free callno's - so %d calls are active", i, CALLNO_TOTAL - i);

	bzero(used, sizeof used);
	bzero(unused, sizeof unused);

	/* count up all the unused ones */
	list_foreach (&free_callno_list, iter) {
		no = list_entry(iter, free_callno, list)->no;
		if (unused[no])
			CRASH("duplicate unused callno %d", no);
		unused[no] = 1;
	}

	/* count ones used by POKE-ing servers */
	list_foreach(&servers, iter) {
		srv = list_entry(iter, server, list);
		list_foreach(&srv->opening_conns_head, conn_iter) {
			no = list_entry(conn_iter, connection, opening_list)->cltno;
			if (used[no])
				CRASH("duplicate used callno %d", no);
			if (unused[no])
				CRASH("callno %d is both used AND unused??", no);
			used[no] = 1;
		}
	}

	/* count ones used by client connections */
	for (i = 0; i < 256; i++) {
		list_foreach (&clients[i], iter) {
			for (j = 0; j < CALLNO_SIZE; j++) {
				conn = list_entry(iter, client, list)->connections[j];
				if (conn) {
					no = conn->balancer_cltno;
					if (used[no])
						CRASH("duplicate used callno %d", no);
					if (unused[no])
						CRASH("callno %d is both used AND unused??", no);
					used[no] = 1;
				}
			}
		}
	}

	/* callnos used on pending ACKs from client POKEs */
	list_foreach(&ack_wait_list, iter) {
		no = list_entry(iter, ack_wait, list)->srvno;
		if (used[no])
			CRASH("duplicate used callno %d", no);
		if (unused[no])
			CRASH("callno %d is both used AND unused??", no);
		used[no] = 1;
	}

	for (i = CALLNO_MIN; i <= CALLNO_MAX; i++) {
		if (!used[i] && !unused[i])
			CRASH("callno %d is lost - neither used nor unused", i);
	}
}

int32_t free_callno_get(void) {
	list_ptrs *iter;

	iter = free_callno_list.next;
	if(iter!=NULL) {
		list_remove(iter);
		return (int32_t) (list_entry(iter, free_callno, list)->no);
	} else {
		LOG(ERROR, "*** RUN OUT OF CALLNOs ***");
		return -1;
	}
}

void free_callno_return(uint16_t no) {
	free_callno *f;

	f = &free_callnos[no];
	if (f->list.next || f->list.prev)
		CRASH("returning callno %d which is already in free list", no);

	list_add_after(&free_callno_list, &free_callnos[no].list);
	free_callno_count(); // will report errors
}

static void try_retransmit(send_command_t send_command, int sock_fd, retransmission *ret) {
#define GET_ADDR(r) ((r->direction==RETR_DEST_CLIENT)?&r->destination.client->addr:&r->destination.server->addr)
	if(send_command(sock_fd, ret->data, (socklen_t) ret->len, 0, (struct sockaddr*) GET_ADDR(ret), sizeof(struct sockaddr))==-1) {
		LOG(ERROR, "sendto() returned %s", strerror(errno));
	}
	ret->no_tries++;
#undef GET_ADDR
}

static int usec_time_diff(struct timeval *old, struct timeval *new) {
	return (new->tv_sec - old->tv_sec) * 1000000 + (new->tv_usec - old->tv_usec);
}

static retransmission* retransmission_find(RETR_DEST direction, void *destination, struct connection *conn, int len, const char *data) {
	list_ptrs *iter;
	retransmission *ret;
	const struct iax_packet *pkt_new, *pkt_orig;

	list_foreach(&conn->retransmissions_list, iter) {
		ret = list_entry(iter, retransmission, conn_list);
		if (ret->direction != direction)
			continue;

		if (((ret->direction == RETR_DEST_CLIENT)?(void*)ret->destination.client:(void*)ret->destination.server) != destination)
			continue;

		if (ret->len != len)
			continue;

		pkt_new = (const struct iax_packet*) &ret->data;
		pkt_orig = (const struct iax_packet*) data;
		if (pkt_new->iseqno != pkt_orig->iseqno)
			continue;
		if (pkt_new->oseqno != pkt_orig->oseqno)
			continue;

		return ret;
	}

	return NULL;
}

retransmission* retransmission_new(RETR_DEST direction, void *destination, struct connection *conn, int len, const char *data, struct timeval *now) {
	retransmission *ret;
	
	ret = retransmission_find(direction, destination, conn, len, data);
	if(ret)
		return ret;

	ret = malloc(sizeof(retransmission) + len);
	ret->direction = direction;

	if (direction==RETR_DEST_CLIENT) {
		ret->destination.client = (client*) destination;
		client_ref_up((client*) destination);
	} else {
		ret->destination.server = (server*) destination;
	}

	ret->no_tries = 0;
	ret->last_try = *now;
	ret->len = len;

	memcpy(&ret->data, data, len);
	iax_set_retransmission(((struct iax_packet*) &ret->data), 1);

	list_add_after(&retransmissions_list, &ret->list);
	list_add_after(&conn->retransmissions_list, &ret->conn_list);

	return ret;
}

void retransmission_free(retransmission *ret) {
	list_remove(&ret->list);
	list_remove(&ret->conn_list);

	if (ret->direction == RETR_DEST_CLIENT)
		client_ref_down(ret->destination.client);
	free(ret);
}

int retransmission_count(void) {
	int cnt = 0;
	list_ptrs *iter;
	
	list_foreach(&retransmissions_list, iter)
		cnt++;

	return cnt;
}

int retransmission_count_connection(connection *conn) {
	int cnt = 0;
	list_ptrs *iter;
	
	list_foreach(&conn->retransmissions_list, iter)
		cnt++;

	return cnt;
}

void retransmission_clear(RETR_DEST dest, uint8_t seq, list_ptrs *conn_retr_list) {
	list_ptrs *iter, *iter_next;
	retransmission *ret;

	LOG(VERBOSE_DEBUG, "trying to remove retransmission with oseqno = %i", seq);
	list_foreach_safe(conn_retr_list, iter_next, iter) {
		ret = list_entry(iter, retransmission, conn_list);
		
		LOG(VERBOSE_DEBUG, "comparing with retransmission of \"%s\", oseqno = %i", FRAME_NAME(ret->data), (ret->data)->oseqno);
		log_packet(VERBOSE_DEBUG, ret->data);
		if ((ret->data)->oseqno == seq && ret->direction == dest) {
			retransmission_free(ret);
			
			// there can be only one packet with a specific oseqno
			return;
		}
	}
}

void handle_retransmissions(send_command_t send_command, int sock_fd, struct timeval *now) {
	list_ptrs *iter, *iter_next;
	retransmission *ret;
	
	list_foreach_safe(&retransmissions_list, iter_next, iter) {
		ret = list_entry(iter, retransmission, list);

		if (usec_time_diff(&ret->last_try, now) > RETRANSMISSION_USEC) {
			try_retransmit(send_command, sock_fd, ret);
			if (ret->no_tries > RETRANSMISSION_TRIES)
				retransmission_free(ret);
			else
				ret->last_try = *now;
		}
	}
}

