/*
IAX balancing proxy
Authors: Ben Smithrust, Stanislaw Pitucha

Copyright (c) 2007 Gradwell dot com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#ifndef _CONNECTIONS_H
#define	_CONNECTIONS_H

// retransmission destinations
typedef enum {
	RETR_DEST_SERVER,
	RETR_DEST_CLIENT
} RETR_DEST;

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

typedef struct connection connection;
typedef ssize_t (*send_command_t)(int, const void *, size_t, int, const struct sockaddr *, socklen_t);

#include "clients.h"
#include "global.h"
#include "servers.h"

struct connection {
	server *srv;
	/*@null@*/ client *clt;
	uint16_t cltno, srvno, balancer_cltno; // callno
	int is_closing:1;
	uint32_t closing, setup_timestamp; // timestamp of closing message
	list_ptrs opening_list, global_list;
	list_ptrs retransmissions_list;
	time_t created, last_packet;

	char *from, *to, *user;
	int from_len, to_len, user_len;
};

typedef struct ack_wait {
	uint16_t cltno, srvno;
	client *clt;
	list_ptrs list;
	time_t last_packet;
	uint32_t ts;
} ack_wait;

typedef struct free_callno {
	uint16_t no;
	list_ptrs list;
} free_callno;

#include "iax.h"

typedef struct retransmission {
	list_ptrs list;
	list_ptrs conn_list;

	RETR_DEST direction;
	union {
		client *client;
		server *server;
	} destination;

	int no_tries;
	struct timeval last_try;
	
	int len;
	struct iax_packet data[0];
} retransmission;

#define ADDRESS_FULL_TO_SERVER(frame, conn) do { frame->scallno = htons(conn->balancer_cltno | 0x8000); frame->dcallno = htons(conn->srvno & 0x7fff); } while(0)
#define ADDRESS_MINI_TO_SERVER(frame, conn) do { frame->scallno = htons(conn->balancer_cltno & 0x7fff); } while(0)
#define ADDRESS_FULL_TO_CLIENT(frame, conn) do { frame->scallno = htons(conn->balancer_cltno | 0x8000); frame->dcallno = htons(conn->cltno & 0x7fff); } while(0)
#define ADDRESS_MINI_TO_CLIENT(frame, conn) do { frame->scallno = htons(conn->balancer_cltno & 0x7fff); } while(0)

connection* connection_create_from_balancer(/*@only@*/ server *srv);
/*@null@*/ connection* connection_get_by_clt(client *clt, uint16_t cltno, uint16_t srvno, char create);
/*@null@*/ connection* connection_get_by_srv(server *srv, uint16_t cltno, uint16_t srvno);
void connection_cleanup_check(connection *conn, time_t now);
void connection_cleanup_lists(struct timeval *now);
void connection_print(connection *conn);
void connection_list_print(void);
int connection_check_if_exist(void);
void connection_destroy_all(void);
int connection_count(void);

int ack_wait_lookup(client *clt, uint16_t cltno, uint16_t srvno, uint32_t ts); // return 1 if ack is for balancer + unlink, 0 otherwise
void ack_wait_add(client *clt, uint16_t cltno, uint16_t srvno, uint32_t ts);

void free_callno_init(void);
int32_t free_callno_get(void);
void free_callno_return(uint16_t);
void free_callno_log(void);

#define retransmission_new_client(clt, conn, len, data, now) retransmission_new(RETR_DEST_CLIENT, clt, conn, len, data, now)
#define retransmission_new_server(clt, conn, len, data, now) retransmission_new(RETR_DEST_SERVER, clt, conn, len, data, now)
retransmission* retransmission_new(RETR_DEST, void *, struct connection*, int, const char *, struct timeval *);
void retransmission_free(retransmission *);
int retransmission_count(void);
int retransmission_count_connection(connection *conn);
void handle_retransmissions(send_command_t, int, struct timeval*);
void retransmission_clear(RETR_DEST, uint8_t, list_ptrs *);

#endif	/* _CONNECTIONS_H */

