#include <stdlib.h>
#include <check.h>
#include "../src/clients.h"

START_TEST(test_client_get)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);
	client *clt = client_get(address, 1234);

	fail_if(clt == NULL, "client not created");

	client_destroy(clt);
}
END_TEST

START_TEST(test_client_search_fail)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client_search cs = client_find_and_count(address, 1234);
	fail_if(cs.found, "client should not be present");
	fail_unless(cs.result.ip_count == 0, "there should be no clients (found %d)", cs.result.ip_count);
}
END_TEST

START_TEST(test_client_search_others)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client_get(address, 1234);
	client_get(address, 5678);

	client_search cs = client_find_and_count(address, 9999);
	fail_if(cs.found, "client should not be present");
	fail_unless(cs.result.ip_count == 2, "there should be 2 clients (found %d)", cs.result.ip_count);

	client_destroy_all();
}
END_TEST

START_TEST(test_client_search_found)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client_get(address, 1111);
	client *clt = client_get(address, 2222);
	client_get(address, 3333);

	client_search cs = client_find_and_count(address, 2222);
	fail_unless(cs.found, "client should be present");
	fail_unless(cs.result.clt == clt, "wrong client returned");

	client_destroy_all();
}
END_TEST

START_TEST(test_client_doubleget)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client *clt1 = client_get(address, 1234);
	fail_if(clt1 == NULL, "client 1 cannot be NULL");
	client *clt2 = client_get(address, 1234);
	fail_if(clt1 == NULL, "client 2 cannot be NULL");

	fail_unless(clt1 == clt2, "client not found");
	
	client_destroy(clt1);
}
END_TEST

START_TEST(test_client_diffport)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client *clt1 = client_get(address, 1234);
	fail_if(clt1 == NULL, "client 1 cannot be NULL");
	client *clt2 = client_get(address, 5678);
	fail_if(clt2 == NULL, "client 2 cannot be NULL");

	fail_if(clt1 == clt2, "client should be different");
	
	client_destroy(clt1);
	client_destroy(clt2);
}
END_TEST

START_TEST(test_client_refcount)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	client *clt = client_get(address, 1234);
	fail_unless(clt->ref_count == 0, "number of references should be 0");
	
	client_ref_up(clt);
	fail_unless(clt->ref_count == 1, "number of references should be 1");
	
	client_ref_down(clt);
	fail_unless(clt->ref_count == 0, "number of references should be 0");
	
	client_destroy(clt);
}
END_TEST

START_TEST(test_client_refdestroy)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	fail_unless(client_count() == 0, "number of clients should be 0");
	client *clt = client_get(address, 1234);
	
	fail_unless(client_count() == 1, "number of clients should be 1");
	
	client_ref_up(clt);
	client_ref_down_destroy(clt);

	fail_unless(client_count() == 0, "number of clients should be 0");
}
END_TEST

START_TEST(test_client_destroyall)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	fail_unless(client_count() == 0, "number of clients should be 0");
	client *clt = client_get(address, 1234);
	
	fail_unless(client_count() == 1, "number of clients should be 1");
	
	client_destroy_all();
	fail_unless(client_count() == 0, "number of clients should be 0");
}
END_TEST

START_TEST(test_client_getcallno)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);

	free_callno_init();
	client *clt = client_get(address, 1234);
	
	fail_unless(clt->callno_count == 0, "client should not have callnos assigned (has %d)", clt->callno_count);
	int callno = client_get_callno(clt);
	fail_unless(callno > 0 && callno < 0x8000, "couldn't assign a correct callno (got %d)", callno);

	client_destroy_all();
}
END_TEST

START_TEST(test_client_limitcallno)
{
	struct in_addr address;
	inet_aton("1.2.3.4", &address);
	int i, callnos[MAX_CALLNO_PER_CLT+1];

	free_callno_init();
	client *clt = client_get(address, 1234);
	
	for (i=0; i<MAX_CALLNO_PER_CLT; i++) {
		callnos[i] = client_get_callno(clt);
		fail_unless(callnos[i] > 0 && callnos[i] < 0x8000, "couldn't assign a correct callno in round %d (got %d)", i, callnos[i]);
	}
	
	callnos[i] = client_get_callno(clt);
	fail_unless(callnos[i] == -1, "callno should not be assigned (over the limit)");

	client_destroy_all();
}
END_TEST

Suite *clients_suite (void) {
	Suite *s = suite_create ("Clients");

	/* Core test case */
	TCase *tc_counting = tcase_create ("counting");
	tcase_add_test (tc_counting, test_client_search_fail);
	tcase_add_test (tc_counting, test_client_search_others);
	tcase_add_test (tc_counting, test_client_search_found);
	suite_add_tcase (s, tc_counting);

	TCase *tc_creating = tcase_create ("creating");
	tcase_add_test (tc_creating, test_client_get);
	tcase_add_test (tc_creating, test_client_doubleget);
	tcase_add_test (tc_creating, test_client_diffport);
	suite_add_tcase (s, tc_creating);
	
	TCase *tc_ref = tcase_create ("references");
	tcase_add_test (tc_ref, test_client_refcount);
	tcase_add_test (tc_ref, test_client_refdestroy);
	tcase_add_test (tc_ref, test_client_destroyall);
	suite_add_tcase (s, tc_ref);
	
	TCase *tc_callnos = tcase_create ("callnos");
	tcase_add_test (tc_callnos, test_client_getcallno);
	tcase_add_test (tc_callnos, test_client_limitcallno);
	suite_add_tcase (s, tc_callnos);

	return s;
}
		
int main (void) {
	int number_failed;
	Suite *s = clients_suite ();
	SRunner *sr = srunner_create (s);
	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);
	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}


